#!/usr/bin/env python
#
# Freely rotating chain (FRC) simulation following Livadaru et al. [1].
#
# [1]: http://pubs.acs.org/doi/abs/10.1021/ma020751g

import numpy as _n
import matplotlib.pyplot as _pyplot
from mpl_toolkits.mplot3d import Axes3D as _Axes3D
try:  # SciPy >= 0.10.0
    from scipy.constants import Boltzmann as _kB
except ImportError:  # SciPy < 0.10.0
    from scipy.constants import Bolzmann as _kB
from scipy.integrate import quad as _quad


MANUAL_T_INTEGRAL = True
MANUAL_GIP_INTEGRAL = True
MANUAL_GIBBS_INTEGRAL = True
MANUAL_INTEGRAL_STEPS = 5000
_T_cache = {}


def _integral(func, a, b, args, manual=True, M=MANUAL_INTEGRAL_STEPS):
    if manual:
        dx = (b-a)/(M-1)
        integral = 0
        for x in _n.linspace(a, b, M):
            xp = x + dx
            y = func(x, *args)
            integral += y*dx
    else:
        integral,abserr = _quad(func, a, b, args=args)
    return integral


def _Gip_kernel(theta_prime, theta, Gis, fbkBT, gamma, b):
    "eqn. 24, kernel"
    transfer = T(theta=theta, theta_prime=theta_prime, gamma=gamma, b=b)
    if len(Gis) == 1:  # only holding G0
        return Gis[-1](theta=theta, fbkBT=fbkBT) * transfer
    else:  # holding Gi>0
        return Gip(
            theta=theta_prime, Gis=Gis[:-1], fbkBT=fbkBT,
            gamma=gamma, b=b) * transfer

def Gip(theta, Gis, fbkBT, gamma, b):
    "eqn. 24 wrapper"
    integral = _integral(
        func=_Gip_kernel, a=0, b=2*_n.pi, args=(theta, Gis, fbkBT, gamma, b),
        manual=MANUAL_GIP_INTEGRAL)
    return _n.exp(fbkBT*_n.cos(theta)) * abs(integral)

def delta(y, y_p):
    if (y_p == 0 and y == 0):  # double, but we'll get a single too.
        return 1
    elif y_p == 0:  # single whammy
        return 1
    elif y/y_p < 0:  # crossing
        return 1
    return 0

def delta_h(theta, theta_prime, phi, phi_p, gamma):
    y = theta_prime - h(theta=theta, phi=phi, gamma=gamma)
    y_p = theta_prime - h(theta=theta, phi=phi_p, gamma=gamma)
    return delta(y=y, y_p=y_p)

def T(theta, theta_prime, gamma, b):
    "eqn. 25"
    key = (theta, theta_prime, gamma)
    if key not in _T_cache:
        assert MANUAL_T_INTEGRAL, MANUAL_T_INTEGRAL
        #integral,abserr = _quad(
        #    _T_kernel, 0, 2*_n.pi, args=(theta, theta_prime, gamma))
        M = MANUAL_INTEGRAL_STEPS
        d_phi = 2*_n.pi/(M-1)
        integral = 0
        for phi in _n.linspace(0, 2*_n.pi, M):
            phi_p = phi + d_phi
            integral += delta_h(
                theta=theta, theta_prime=theta_prime, phi=phi, phi_p=phi_p,
                gamma=gamma)
        _T_cache[key] = integral
    return _T_cache[key]

def _h(theta, phi, gamma):
    "for eqn.s 26 and 31"
    return _n.cos(theta)*_n.cos(gamma) + _n.sin(theta)*_n.sin(gamma)*_n.cos(phi)

def h(theta, phi, gamma):
    "eqn. 26"
    return _n.arccos(_h(theta=theta, phi=phi, gamma=gamma))

def G0(theta, fbkBT):
    "eqn. 27"
    return _n.exp(fbkBT*_n.cos(theta)) * abs(_n.sin(theta))

def GN(theta, N, fbkBT, gamma, b=None):
    "for fig. 4"
    Gis = [G0] + [Gip]*(N-1)
    if N == 1:
        return Gis[-1](theta=theta, fbkBT=fbkBT)
    return Gis[-1](theta=theta, Gis=Gis[:-1], fbkBT=fbkBT, gamma=gamma, b=b)

def monte_carlo_GN(N=1, fbkBT=1, gamma_degrees=20, b=1, plot=True):
    "a check for GN()"
    gamma = gamma_degrees * _n.pi / 180
    dirs = _n.zeros((N, 3), dtype=float)
    pos = _n.zeros((N+1, 3), dtype=float)
    for i in range(N):
        dice = 1
        p = 0
        hits = 0
        misses = 0
        scale = 0.001
        while dice > p:
            misses += 1
            if i == 0:
                new_direction = _n.random.randn(3)
                new_direction /= _n.linalg.norm(new_direction)
            else:
                old_direction = dirs[i-1,:]
                phi = 2 * _n.pi * _n.random.rand()
                if old_direction[0] == 0:
                    basis = _n.array([1, 0, 0], dtype=float)
                else:
                    basis = _n.array([0, 1, 0], dtype=float)
                perpendicular_1 = _n.cross(old_direction, basis)
                perpendicular_1 /= _n.linalg.norm(perpendicular_1)
                perpendicular_2 = _n.cross(old_direction, perpendicular_1)
                new_direction = (
                    old_direction * _n.cos(gamma) +
                    perpendicular_1 * _n.sin(gamma) * _n.cos(phi) +
                    perpendicular_2 * _n.sin(gamma) * _n.sin(phi))
            if float(hits)/misses > 0.01:
                scale /= 2
            elif float(hits)/misses < 0.001:
                scale *= 2
            p = _n.exp(fbkBT*new_direction[0]) * scale
            dice = _n.random.rand()
        misses -= 1
        hits += 1
        dirs[i,:] = new_direction
        pos[i+1,:] = pos[i,:] + new_direction
    print('scale: {}'.format(scale))
    if plot:
        figure = _pyplot.figure()
        axes = figure.add_subplot(1, 1, 1, projection='3d', aspect='equal')
        axes.plot([x for x,y,z in pos], [y for x,y,z in pos], [z for x,y,z in pos])
        axes.set_title('monte carlo FRC (gamma={} deg)'.format(gamma_degrees))
        axes.set_xlabel('x')
        axes.set_ylabel('y')
        axes.set_ylabel('z')
        figure.savefig('mv-frc-g-{}.png'.format(gamma_degrees))
    return _n.arccos(dirs[-1][0])

def Gibbs(N, fbkBT, gamma, b):
    "eqn. 28"
    Gis = [G0] + [Gip]*(N-1)
    integral = _integral(
        Gis[-1], 0, 2*_n.pi, args=(Gis[-1], fbkBT, gamma, b),
        manual=MANUAL_GIBBS_INTEGRAL)

def Rz(N, f, b, kBT, gamma, df):
    "eqn. 29"
    G = Gibbs(N=N, fbkBT=f*b/kBT, gamma=gamma)
    Gp = Gibbs(N=N, fbkBT=(f+df)*b/kBT, gamma=gamma)
    return (Gp-G)/df

def test_h(gamma_degrees=20):
    figure = _pyplot.figure()
    axes = figure.add_subplot(1, 1, 1)
    axes.hold(True)
    gamma = gamma_degrees * _n.pi / 180
    for i in range(100):
        thetas = _n.linspace(0*_n.pi, 2*_n.pi, 50)
        phis = _n.linspace(0, 2*_n.pi, 50)
        data = _n.zeros((len(thetas), len(phis)), dtype=float)
        for c,phi in enumerate(phis):
            for r,theta in enumerate(thetas):
                data[r,c] = h(theta, phi, gamma)
    X,Y = _n.meshgrid(thetas, phis)
    pcolor = axes.pcolor(X, Y, data.T)
    figure.colorbar(pcolor)
    axes.axis([0, 2*_n.pi, 0, 2*_n.pi])
    axes.set_title('h(theta, phi, gamma={} deg)'.format(gamma_degrees))
    axes.set_xlabel('theta (degrees)')
    axes.set_ylabel('phi (degrees)')
    figure.savefig('test-h-g-{}.png'.format(gamma_degrees))

def test_T(gamma_degrees=20):
    figure = _pyplot.figure()
    axes = figure.add_subplot(1, 1, 1)
    axes.hold(True)
    gamma = gamma_degrees * _n.pi / 180
    for i in range(100):
        thetas = _n.linspace(0*_n.pi, 2*_n.pi, 50)
        theta_primes = _n.linspace(0, 2*_n.pi, 50)
        data = _n.zeros((len(thetas), len(theta_primes)), dtype=float)
        for c,theta_prime in enumerate(theta_primes):
            for r,theta in enumerate(thetas):
                data[r,c] = T(theta, theta_prime, gamma)
    X,Y = _n.meshgrid(thetas, theta_primes)
    pcolor = axes.pcolor(X, Y, data.T)
    figure.colorbar(pcolor)
    axes.axis([0, 2*_n.pi, 0, 2*_n.pi])
    axes.set_title("T(theta, theta', gamma={} deg)".format(gamma_degrees))
    axes.set_xlabel('theta (degrees)')
    axes.set_ylabel("theta' (degrees)")
    figure.savefig('test-T-g-{}.png'.format(gamma_degrees))

def figure_4_frame(N=3, gamma_degrees=20):
    figure = _pyplot.figure()
    axes = figure.add_subplot(1, 1, 1)
    axes.hold(True)
    gamma = gamma_degrees * _n.pi / 180
    for color,N in [('r', 1), ('b', 2)]: #, ('g', 3)]:
        fbkBTs = [100] #_n.linspace(0, 100, 5)
        thetas = _n.linspace(0, 2*_n.pi, 50)
        data = _n.zeros((len(thetas), len(fbkBTs)), dtype=float)
        for c,fbkBT in enumerate(fbkBTs):
            for r,theta in enumerate(thetas):
                print((r,c))
                #data[r,c] = G0(theta=theta, fbkBT=fbkBT)
                data[r,c] = GN(theta=theta, N=N, fbkBT=fbkBT, gamma=gamma)
            data[:,c] /= data[:,c].sum()
        degrees = thetas * 180 / _n.pi
        axes.plot(degrees, data, '{}.-'.format(color))
    axes.axes.axvline(gamma_degrees/2., color='k')
    axes.set_xbound([0, 100])
    axes.set_title('fig. 4 (gamma = {} deg)'.format(gamma_degrees))
    axes.set_xlabel('theta (degrees)')
    axes.set_ylabel('probability G_N(theta)')
    figure.savefig('test-G_N-g-{}.png'.format(gamma_degrees))

###
#
#def hz(theta, phi, gamma, b):
#    "eqn. 31"
#    return b * _h(theta=theta, phi=phi, gamma=gamma)
#
#def delta_hz(theta, z, z_prime, phi, phi_p, gamma, b):
#    "for eqn. 30"
#    y = z_prime - z - hz(theta=theta, phi=phi, gamma=gamma)
#    y_p = z_prime - hz(theta=theta, phi=phi_p, gamma=gamma)
#    return delta(y=y, y_p=y_p)
#
#def Tz(z, theta, z_prime, theta_prime, gamma, b):
#    "eqn. 30"
#    key = (z, theta, z_prime, theta_prime, gamma, b)
#    if key not in _T_cache:
#        assert MANUAL_T_INTEGRAL, MANUAL_T_INTEGRAL
#        #integral,abserr = _quad(
#        #    _T_kernel, 0, 2*_n.pi, args=(theta, theta_prime, gamma, b))
#        M = MANUAL_INTEGRAL_STEPS
#        d_phi = 2*_n.pi/(M-1)
#        integral = 0
#        for phi in _n.linspace(0, 2*_n.pi, M):
#            phi_p = phi + d_phi
#            integral += delta_h(
#                theta=theta, theta_prime=theta_prime, phi=phi, phi_p=phi_p,
#                gamma=gamma) * delta_hz(
#                theta=theta, z=z, z_prime=z_prime, phi=phi, phi_p=phi_p,
#                gamma=gamma, b)
#        _T_cache[key] = integral
#    return _T_cache[key]
#
#def G0Z(z, theta, gamma, b):
#    return delta(z-b*_n.cos(theta))*_n.sin(theta)



if __name__ == '__main__':
    for gamma_degrees in [20, 60]: #, 90]:
        #test_h(gamma_degrees=gamma_degrees)
        #test_T(gamma_degrees=gamma_degrees)
        figure_4_frame(N=100, gamma_degrees=gamma_degrees)
        #thetas = []
        #for x in range(500):
        #    print('calc theta {} for {}'.format(x, gamma_degrees))
        #    thetas.append(monte_carlo_GN(N=100, fbkBT=100, gamma_degrees=gamma_degrees, plot=False))
        #figure = _pyplot.figure()
        #axes = figure.add_subplot(1, 1, 1)
        #axes.hist(_n.array(thetas) * 180 / _n.pi)
        #print('done hist')
    _pyplot.show()
